"""
Auto-generated scaffold for agent-creation-copilot.
Modify AGENT_DESCRIPTION and configuration fields as needed.
"""

from __future__ import annotations

from typing import Any

from agenthub.interfaces.scaffolds.autonomous_agent import (
    AutonomousAgentData,
    IAgentInstructionDTO,
)


AGENT_NAME = 'agent-creation-copilot'
AGENT_DESCRIPTION = 'agent-creation-copilot agent description'

AGENT_ID = 4  # Not defined if not synchronized with AgentHub.
MODULE_ID = 'autonomous-agent'
LLM_STRING = 'openai:gpt-4o'
PROMPT_FILENAME = 'agent_creation_copilot.prompt'


def _coerce_prompt(prompt_text: str | None) -> IAgentInstructionDTO | None:
    if not prompt_text:
        return None

    cleaned = "\n".join(
        line for line in prompt_text.splitlines()
        if not line.lstrip().startswith("//")
    )
    return IAgentInstructionDTO.from_yaml(cleaned)


def get_agent_manifest(prompt: str | None = None) -> dict[str, Any]:
    configuration = get_agent_configuration(prompt)
    return {
        "name": AGENT_NAME,
        "avatar_url": None,
        "description": AGENT_DESCRIPTION,
        "module_id": MODULE_ID,
        "configuration": configuration,
    }


def get_agent_configuration(prompt: str | None = None) -> AutonomousAgentData:
    instructions = _coerce_prompt(prompt) or IAgentInstructionDTO(
        objective='You are an AgentHub agent. Fall back to this instruction when no prompt is provided.',
        role=None,
        backstory=None,
        constraints=[],
        preferences=[],
        output_format="json",
    )
    return AutonomousAgentData(
        llm=LLM_STRING,
        instructions=instructions,
        policies=["think_before_acting", "see_own_tokens"],
        attp_catalogs=["agenthub-copilot"],
        # enable_attp=False,
        # enable_mcp=False
    )



from __future__ import annotations

import importlib
import importlib.resources as resources
import importlib.util
import inspect
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from agenthub.interfaces.scaffold_module import ScaffoldModule


def _slugify(value: str) -> str:
    slug = re.sub(r"[^a-z0-9]+", "_", value.lower()).strip("_")
    return slug or "agent"


def _python_repr(value: str | None) -> str:
    return "None" if value is None else repr(value)


@dataclass(slots=True)
class ScaffoldGenerationSpec:
    name: str
    description: str | None = None
    module_id: str = "common-agent"
    agent_id: int | None = None
    llm_load_string: str = "openai:gpt-5"
    instruction_template: str | None = None
    avatar_url: str | None = None
    attp_catalog: str | None = None
    prompt_text: str | None = None
    overwrite: bool = False
    scaffolds_dir: Path | None = None


class ScaffoldEngine:
    def __init__(self, scaffolds_dir: Path | None = None) -> None:
        self.scaffolds_dir = scaffolds_dir or (Path.cwd() / "scaffolds")

    def discover_scaffold_modules(self) -> dict[str, ScaffoldModule]:
        modules: dict[str, ScaffoldModule] = {}

        def _load_module(module_name: str) -> None:
            try:
                mod = importlib.import_module(module_name)
            except Exception:
                return
            for obj in mod.__dict__.values():
                if (
                    inspect.isclass(obj)
                    and issubclass(obj, ScaffoldModule)
                    and obj is not ScaffoldModule
                ):
                    instance = obj()
                    if instance.module_id and instance.module_id not in modules:
                        modules[instance.module_id] = instance

        # Try packaged resources first
        try:
            pkg_files = resources.files("agenthub.scaffolds")
            for entry in pkg_files.iterdir():
                if entry.name.endswith(".py") and entry.name != "__init__.py":
                    _load_module(f"agenthub.scaffolds.{entry.name[:-3]}")
        except Exception:
            pass

        # Fallback to local filesystem if nothing found
        if not modules:
            local_dir = Path(__file__).resolve().parents[2] / "scaffolds"
            for path in local_dir.glob("*.py"):
                if path.name == "__init__.py":
                    continue
                module_name = f"agenthub.scaffolds.{path.stem}"
                spec = importlib.util.spec_from_file_location(module_name, path)
                if not spec or not spec.loader:
                    continue
                mod = importlib.util.module_from_spec(spec)
                sys.modules[module_name] = mod
                try:
                    spec.loader.exec_module(mod)
                except Exception:
                    continue
                for obj in mod.__dict__.values():
                    if (
                        inspect.isclass(obj)
                        and issubclass(obj, ScaffoldModule)
                        and obj is not ScaffoldModule
                    ):
                        instance = obj()
                        if instance.module_id and instance.module_id not in modules:
                            modules[instance.module_id] = instance

        return modules

    def generate_scaffold(self, spec: ScaffoldGenerationSpec) -> dict[str, Any]:
        modules = self.discover_scaffold_modules()
        if not modules:
            raise RuntimeError("No scaffold modules available.")

        module = modules.get(spec.module_id)
        if not module:
            raise RuntimeError(
                f"Unknown module_id '{spec.module_id}'. Available: {list(modules.keys())}"
            )

        slug = _slugify(spec.name)
        scaffolds_dir = self.scaffolds_dir
        scaffold_path = scaffolds_dir / f"{slug}.py"
        prompt_path = scaffold_path.with_suffix(".prompt")

        if scaffold_path.exists() and not spec.overwrite:
            raise FileExistsError(
                f"Scaffold {scaffold_path} exists. Set overwrite=True to replace."
            )

        description = spec.description or f"{spec.name} agent description"
        instruction_template = spec.instruction_template or (
            "You are an AgentHub agent. Fall back to this instruction when no prompt is provided."
        )

        context = {
            "agent_name": spec.name,
            "agent_name_repr": repr(spec.name),
            "agent_id": spec.agent_id,
            "description_repr": repr(description),
            "module_id": spec.module_id,
            "module_id_repr": repr(spec.module_id),
            "llm_load_string_repr": repr(spec.llm_load_string),
            "prompt_filename_repr": repr(prompt_path.name),
            "instruction_template_repr": repr(instruction_template),
            "avatar_url_repr": _python_repr(spec.avatar_url),
            "attp_catalog_repr": _python_repr(spec.attp_catalog),
            "enable_attp": True,
            "enable_tables_reading": False,
        }

        prompt_default = module.make_prompt_template(context)
        prompt_content = spec.prompt_text or prompt_default
        scaffold_content = module.generate_manifest(context)

        scaffolds_dir.mkdir(parents=True, exist_ok=True)
        scaffold_path.write_text(scaffold_content.rstrip() + "\n", encoding="utf-8")
        prompt_path.write_text(prompt_content.rstrip() + "\n", encoding="utf-8")

        return {
            "slug": slug,
            "module_id": module.module_id,
            "scaffold_path": str(scaffold_path),
            "prompt_path": str(prompt_path),
        }



from __future__ import annotations

from typing import Any

from ascender.core import Provider
from ascender.core.cli_engine import useCLI

from agenthub.cli.agents import AgentBuilderCLI
from agenthub.cli.copilot import CopilotCLI
from agenthub.configs.main import AgentHubConfig, load_agenthub_config


def provideAgentHub() -> Provider:
    config: AgentHubConfig = load_agenthub_config()
    return [
        {
            "provide": "AGENTHUB_CONFIG",
            "value": config,
        }, 
        useCLI(AgentBuilderCLI), 
        useCLI(CopilotCLI)
    ]



from .attp_tool import AttpTool
from .context import Ctx
from .pipe import Pipe
from .standalone_tool import StandaloneTool, tool

__all__ = [
    "AttpTool",
    "Ctx",
    "Pipe",
    "StandaloneTool",
    "tool",
]



from typing import Any

from ascender.core import inject
from attp_client import ATTPClient, AttpException, FixedBaseModel
from attp_client.interfaces.objects.agent import IAgentDTO, IAgentResponse
from pydantic import Field
import requests
import re

from agenthub.attp.standalone_tool import tool
from agenthub.configs.main import AgentHubConfig
from agenthub.interfaces.scaffolds.autonomous_agent import AutonomousAgentData
from agenthub.interfaces.scaffolds.common_agent import CommonAgentData
from agenthub.core.scaffolds import ScaffoldEngine, ScaffoldGenerationSpec


def _slugify(value: str) -> str:
    slug = re.sub(r"[^a-z0-9]+", "_", value.lower()).strip("_")
    return slug or "agent"


def _python_repr(value: str | None) -> str:
    return "None" if value is None else repr(value)


class ConfigSchemaRequest(FixedBaseModel):
    module_id: str


class CreateAgentRequest(FixedBaseModel):
    name: str
    description: str
    module_id: str
    configuration: dict[str, Any]
    avatar_url: str | None = None


class ValidateConfigurationRequest(FixedBaseModel):
    module_id: str
    configuration: dict[str, Any] = Field(description="An agent config you want to validate before creating.")


class ListAgentsRequest(FixedBaseModel):
    search: str | None = None
    page: int
    size: int = 10


class ModifyAgentRequest(FixedBaseModel):
    agent_id: int
    configuration: dict[str, Any]
    name: str | None = None
    description: str | None = None
    avatar_url: str | None = None


class DeleteAgentRequest(FixedBaseModel):
    agent_id: int


class GenerateScaffoldRequest(FixedBaseModel):
    name: str
    description: str | None = None
    module_id: str = "common-agent"
    agent_id: int | None = None
    llm_load_string: str = "openai:gpt-5"
    instruction_template: str | None = None
    avatar_url: str | None = None
    attp_catalog: str | None = None
    prompt_text: str | None = None
    overwrite: bool = False


class CopilotTools:
    def __init__(self, client: ATTPClient) -> None:
        self.client = client
        self.scaffold_engine = ScaffoldEngine()

    async def attach_all(self) -> None:
        await self.config_schema.attach(self.client)
        await self.create_agent.attach(self.client)
        await self.validate_configuration.attach(self.client)
        await self.list_agents.attach(self.client)
        await self.modify_agent.attach(self.client)
        await self.delete_agent.attach(self.client)
        await self.generate_scaffold.attach(self.client)

    @tool("generate_scaffold", "Generate a scaffold file and prompt for an agent.", catalog="agenthub-copilot")
    def generate_scaffold(self, request: GenerateScaffoldRequest) -> dict[str, Any]:
        try:
            spec = ScaffoldGenerationSpec(
                name=request.name,
                description=request.description,
                module_id=request.module_id,
                agent_id=request.agent_id,
                llm_load_string=request.llm_load_string,
                instruction_template=request.instruction_template,
                avatar_url=request.avatar_url,
                attp_catalog=request.attp_catalog,
                prompt_text=request.prompt_text,
                overwrite=request.overwrite,
            )
            result = self.scaffold_engine.generate_scaffold(spec)
            return result
        except FileExistsError as exc:
            raise AttpException("ScaffoldExists", detail={"message": str(exc)}) from exc
        except RuntimeError as exc:
            raise AttpException("ScaffoldModuleNotFound", detail={"message": str(exc)}) from exc

    @tool("config_schema", "Get the configuration schema for an agent module.", catalog="agenthub-copilot")
    def config_schema(self, request: ConfigSchemaRequest) -> dict[str, Any]:
        print(request.module_id)
        match request.module_id:
            case "autonomous-agent":
                schema = AutonomousAgentData.model_json_schema()
                return schema
            case "common-agent":
                schema = CommonAgentData.model_json_schema()
                return schema
            case _:
                raise AttpException("ModuleIdNotFound", detail={"message": f"Module ID '{request.module_id}' not found."})

    @tool("create_agent", "Create a new agent.", catalog="agenthub-copilot")
    async def create_agent(self, request: CreateAgentRequest) -> IAgentResponse:
        response = await self.client.agents.create_agent(IAgentDTO(
            name=request.name,
            description=request.description,
            module_id=request.module_id,
            configurations=request.configuration,
            avatar_url=request.avatar_url,
        ))
        
        return response

    @tool("validate_configuration", "Validate the configuration for an agent module.", catalog="agenthub-copilot")
    def validate_configuration(self, request: ValidateConfigurationRequest) -> dict[str, Any]:
        match request.module_id:
            case "autonomous-agent":
                try:
                    AutonomousAgentData.model_validate(request.configuration)
                    return {"valid": True}
                except Exception as e:
                    return {"valid": False, "error": str(e)}
            case "common-agent":
                try:
                    CommonAgentData.model_validate(request.configuration)
                    return {"valid": True}
                except Exception as e:
                    return {"valid": False, "error": str(e)}
            case _:
                raise AttpException("ModuleIdNotFound", detail={"message": f"Module ID '{request.module_id}' not found."})

    @tool("list_agents", "List autonomous agents.", catalog="agenthub-copilot")
    def list_agents(self, request: ListAgentsRequest) -> dict[str, Any]:
        configs = inject(AgentHubConfig)
        
        agents = requests.get("{}/agents".format(configs.base_url), params={
            "page": request.page,
            "size": request.size,
        }, headers={
            "Authorization": f"Bearer {configs.agt_key}",
        }).json()
        # response = await self.client.agents.get_agents(search=request.search)
        # agents = [
        #     {
        #         "id": agent.id,
        #         "name": agent.name,
        #         "description": agent.description,
        #         "module_id": agent.module_id,
        #         "avatar_url": agent.avatar_url,
        #     }
        #     for agent in response.items
        # ]
        return agents

    @tool("modify_agent", "Modify an existing autonomous agent.", catalog="agenthub-copilot")
    async def modify_agent(self, request: ModifyAgentRequest) -> IAgentResponse:
        agent = await self.client.agents.get_agent(request.agent_id)
        response = await self.client.agents.update_agent(
            agent_id=request.agent_id,
            data=IAgentDTO(
                name=request.name or agent.name,
                description=request.description or agent.description,
                module_id=agent.module_id,
                configurations=request.configuration,
                avatar_url=request.avatar_url,
            )
        )
        return response

    @tool("delete_agent", "Delete an autonomous agent.", catalog="agenthub-copilot")
    async def delete_agent(self, request: DeleteAgentRequest) -> dict[str, Any]:
        await self.client.agents.delete_agent(request.agent_id)
        return {"deleted": True}


from typing import Any, Callable
from ascender.core import ControllerDecoratorHook, inject

from fastapi_mcp import FastApiMCP
from mcp.types import ToolAnnotations, Icon


class MCPTool(ControllerDecoratorHook):
    server: FastApiMCP | None = inject(FastApiMCP, None)
    
    def __init__(
        self,
        name: str | None = None,
        title: str | None = None,
        description: str | None = None,
        annotations: ToolAnnotations | None = None,
        icons: list[Icon] | None = None,
        meta: dict[str, Any] | None = None,
        structured_output: bool | None = None,
    ) -> None:
        super().__init__()
        self.name = name
        self.title = title
        self.description = description
        self.annotations = annotations
        self.icons = icons
        self.meta = meta
        self.structured_output = structured_output
    
    def on_load(self, callable: Callable[..., Any]):
        if not self.server:
            raise RuntimeError("MCPTool hook requires FastApiMCP server instance.")
        pass


from __future__ import annotations

import dataclasses
import inspect
import traceback
from typing import Annotated, Any, Callable, get_args, get_origin, get_type_hints

from ascender.common import BaseDTO, BaseResponse
from attp_client import ATTPClient, AttpException, FixedBaseModel
from attp_client.interfaces.catalogs.tools.envelope import IEnvelope
from pydantic import BaseModel

from .context import CtxMarker
from .pipe import Pipe
from ..interfaces.types.tool_metadata import ToolMetadata


class StandaloneTool:
    def __init__(
        self,
        name: str,
        description: str | None,
        catalog: str,
        callback: Callable[..., Any],
    ) -> None:
        self.name = name
        self.description = description
        self.catalog = catalog
        self.callback = callback
        self.schema = self._extract_schema(callback)

    def __get__(self, instance: Any, owner: type | None = None) -> "StandaloneTool":
        if instance is None:
            return self
        bound = self.callback.__get__(instance, owner or type(instance))
        return StandaloneTool(
            name=self.name,
            description=self.description,
            catalog=self.catalog,
            callback=bound,
        )

    def _resolve_fixed_model(self, annotation: Any):
        if inspect.isclass(annotation) and issubclass(
            annotation, (FixedBaseModel, BaseModel, BaseDTO, BaseResponse)
        ):
            return annotation

        origin = get_origin(annotation)
        if inspect.isclass(origin) and issubclass(
            origin, (FixedBaseModel, BaseModel, BaseDTO, BaseResponse)
        ):
            return origin

        return None

    def _is_dataclass_type(self, annotation: Any) -> bool:
        return inspect.isclass(annotation) and dataclasses.is_dataclass(annotation)

    def _is_context_param(self, param: inspect.Parameter) -> bool:
        if isinstance(param.default, CtxMarker):
            return True

        annotation = param.annotation
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is Annotated and any(isinstance(arg, CtxMarker) for arg in args[1:]):
            return True

        return False

    def _is_pipe_param(self, param: inspect.Parameter) -> bool:
        annotation = param.annotation
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is Annotated and any(isinstance(arg, Pipe) for arg in args[1:]):
            return True

        return False

    def _get_pipe_param(self, param: inspect.Parameter) -> Pipe | None:
        annotation = param.annotation
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is Annotated:
            for arg in args[1:]:
                if isinstance(arg, Pipe):
                    return arg
        return None

    def _extract_schema(self, callback: Any):
        type_hints = get_type_hints(callback)
        for param in inspect.signature(callback).parameters.values():
            if self._is_context_param(param) or self._is_pipe_param(param):
                continue

            annotation = self._resolve_annotation(param, type_hints)
            model = self._resolve_fixed_model(annotation)
            if model:
                return model.model_json_schema()
            if self._is_dataclass_type(annotation):
                return self._build_schema_from_dataclass(annotation)

        return self._build_schema_from_signature(callback)

    def _map_type_to_schema(self, annotation: Any) -> dict[str, Any]:
        origin = get_origin(annotation) or annotation
        args = get_args(annotation)

        if origin in (str, bytes):
            return {"type": "string"}
        if origin is bool:
            return {"type": "boolean"}
        if origin is int:
            return {"type": "integer"}
        if origin is float:
            return {"type": "number"}
        if origin is dict:
            return {"type": "object"}
        if origin in (list, tuple, set):
            items_schema = self._map_type_to_schema(args[0]) if args else {}
            return {"type": "array", "items": items_schema}
        return {}

    def _build_schema_from_signature(self, callback: Any) -> dict[str, Any] | None:
        sig = inspect.signature(callback)
        type_hints = get_type_hints(callback)
        properties: dict[str, Any] = {}
        required: list[str] = []

        for name, param in sig.parameters.items():
            if name in ("self", "cls"):
                continue

            if self._is_context_param(param) or self._is_pipe_param(param):
                continue

            annotation = self._resolve_annotation(param, type_hints)
            args = get_args(annotation)
            base_annotation = next((arg for arg in args if arg is not type(None)), annotation)

            schema = self._map_type_to_schema(base_annotation)
            properties[name] = schema

            default = param.default
            allows_none = type(None) in args
            if default is inspect._empty and not allows_none:
                required.append(name)

        if not properties:
            return None

        return {
            "type": "object",
            "properties": properties,
            "required": required,
        }

    def _build_schema_from_dataclass(self, model: type) -> dict[str, Any]:
        properties: dict[str, Any] = {}
        required: list[str] = []
        for field in dataclasses.fields(model):
            schema = self._map_type_to_schema(field.type)
            properties[field.name] = schema
            if field.default is dataclasses.MISSING and field.default_factory is dataclasses.MISSING:
                required.append(field.name)
        return {
            "type": "object",
            "properties": properties,
            "required": required,
        }

    def _unwrap_annotation(self, annotation: Any) -> Any:
        origin = get_origin(annotation)
        if origin is Annotated:
            return get_args(annotation)[0]
        if args := get_args(annotation):
            return next((arg for arg in args if arg is not type(None)), annotation)
        return annotation

    def _resolve_annotation(
        self, param: inspect.Parameter, type_hints: dict[str, Any]
    ) -> Any:
        return type_hints.get(param.name, param.annotation)

    async def _build_kwargs_from_envelope(
        self, envelope: IEnvelope, callback: Any
    ) -> dict[str, Any]:
        sig = inspect.signature(callback)
        type_hints = get_type_hints(callback)
        data = envelope.data or {}
        kwargs: dict[str, Any] = {}
        model_params: dict[str, Any] = {}

        for name, param in sig.parameters.items():
            if name in ("self", "cls"):
                continue

            if self._is_context_param(param) or self._is_pipe_param(param):
                continue

            annotation = self._unwrap_annotation(
                self._resolve_annotation(param, type_hints)
            )
            model = self._resolve_fixed_model(annotation)
            if model is None and self._is_dataclass_type(annotation):
                model = annotation
            if model:
                model_params[name] = model

        for name, param in sig.parameters.items():
            if name in ("self", "cls"):
                continue

            if self._is_context_param(param):
                metadata: ToolMetadata = envelope.metadata or {}  # type: ignore
                kwargs[name] = metadata
                continue

            if self._is_pipe_param(param):
                pipe = self._get_pipe_param(param)
                if not pipe:
                    continue
                supply = pipe.supply(envelope)
                kwargs[name] = await supply if inspect.iscoroutine(supply) else supply
                continue

            if name in model_params:
                model = model_params[name]
                if self._is_dataclass_type(model):
                    kwargs[name] = model(**data)
                else:
                    kwargs[name] = model.model_validate(data)
                continue

            if name not in data and param.default is inspect._empty:
                raise ValueError(f"Missing required tool argument '{name}'")

            if name not in data:
                continue

            annotation = self._unwrap_annotation(
                self._resolve_annotation(param, type_hints)
            )
            model = self._resolve_fixed_model(annotation)
            value = data[name]

            if model:
                kwargs[name] = model.model_validate(value)
            elif self._is_dataclass_type(annotation) and isinstance(value, dict):
                kwargs[name] = annotation(**value)
            else:
                kwargs[name] = value

        return kwargs

    async def _invoke(self, envelope: IEnvelope):
        try:
            kwargs = await self._build_kwargs_from_envelope(envelope, self.callback)
            if inspect.iscoroutinefunction(self.callback):
                return await self.callback(**kwargs)
            return self.callback(**kwargs)
        except Exception as exc:
            traceback.print_exc()
            raise AttpException(
                detail={"message": f"Error invoking ATTP tool '{self.name}': {exc}"}
            ) from exc

    async def attach(self, client: ATTPClient) -> None:
        catalog = await client.catalog(self.catalog)
        catalog.attach_tool(
            self._invoke,
            self.name,
            description=self.description or self.callback.__doc__,
            schema=self.schema,
            schema_id=self.name + "_schema",
            schema_ver="2.0",
        )

    async def detach(self, client) -> None:
        catalog = await client.catalog(self.catalog)
        await catalog.detach_tool(self.name)


def tool(
    name: str,
    description: str | None = None,
    *,
    catalog: str = "default",
) -> Callable[[Callable[..., Any]], StandaloneTool]:
    def decorator(func: Callable[..., Any]) -> StandaloneTool:
        return StandaloneTool(name=name, description=description, catalog=catalog, callback=func)

    return decorator



from abc import ABC, abstractmethod
from typing import Any, Callable


class ScaffoldModule(ABC):
    module_id: str = ""
    name: str = ""
    description: str = ""

    @abstractmethod
    def generate_manifest(self, context: dict[str, Any]) -> str:
        """
        Generates a code for scaffold manifest, defining basic values of agent.

        Returns:
            str: A python boilerplate/template code for scaffold manifest.
        """
        raise NotImplementedError("Subclasses must implement generate_manifest method")

    @abstractmethod
    def make_prompt_template(self, context: dict[str, Any]) -> str:
        """
        Either generates a prompt template (if it follows some standard structure or pattern)
        or return a string with default or basic prompt for the module.
        """
        raise NotImplementedError("Subclasses must implement make_prompt_template method")

    @abstractmethod
    def load_scaffold(
        self, loader: Callable[[Any | None], dict[str, Any]], agent_id: int | None = None
    ) -> Any:
        """Loads a scaffold using the provided loader callable and optional agent ID.

        Args:
            loader (Callable[[Any | None], dict[str, Any]]): A callable that takes an optional argument and returns a dictionary of scaffold data.
            agent_id (int | None, optional): An optional agent ID to be used during loading. Defaults to None.

        Returns:
            Any: The loaded scaffold object.
        """
        raise NotImplementedError("Subclasses must implement load_scaffold method")



from __future__ import annotations

from textwrap import dedent
from typing import Any, Callable

from agenthub.interfaces.scaffold_module import ScaffoldModule


class CommonAgentScaffold(ScaffoldModule):
    module_id = "common-agent"
    name = "Common Agent"
    description = "Commonly used agent scaffold."

    def generate_manifest(self, context: dict[str, Any]) -> str:
        agent_id = context.get("agent_id")
        agent_id_repr = "None" if agent_id is None else str(agent_id)
        agent_name_repr = context["agent_name_repr"]
        description_repr = context["description_repr"]
        module_id_repr = context["module_id_repr"]
        llm_load_string_repr = context["llm_load_string_repr"]
        prompt_filename_repr = context["prompt_filename_repr"]
        instruction_template_repr = context["instruction_template_repr"]
        avatar_url_repr = context["avatar_url_repr"]

        return (
            dedent(
                f"""
                \"\"\"
                Auto-generated scaffold for {context['agent_name']}.
                Modify AGENT_DESCRIPTION and configuration fields as needed.
                \"\"\"

                from __future__ import annotations

                from typing import Any

                from agenthub.interfaces.scaffolds.common_agent import CommonAgentData


                AGENT_NAME = {agent_name_repr}
                AGENT_DESCRIPTION = {description_repr}

                AGENT_ID = {agent_id_repr}  # Not defined if not synchronized with AgentHub.
                MODULE_ID = {module_id_repr}
                LLM_STRING = {llm_load_string_repr}
                PROMPT_FILENAME = {prompt_filename_repr}


                def _coerce_prompt(prompt_text: str | None) -> str | None:
                    if not prompt_text:
                        return None

                    return prompt_text.strip()


                def get_agent_manifest(prompt: str | None = None) -> dict[str, Any]:
                    configuration = get_agent_configuration(prompt)
                    return {{
                        \"name\": AGENT_NAME,
                        \"avatar_url\": {avatar_url_repr},
                        \"description\": AGENT_DESCRIPTION,
                        \"module_id\": MODULE_ID,
                        \"configuration\": configuration,
                    }}


                def get_agent_configuration(prompt: str | None = None) -> CommonAgentData:
                    return CommonAgentData(
                        llm=LLM_STRING,
                        instructions=_coerce_prompt(prompt) or "{instruction_template_repr}",
                        enable_attp=True,
                        enable_mcp=False,
                        max_iterations=10
                    )
                """
            ).strip()
            + "\n"
        )

    def make_prompt_template(self, context: dict[str, Any]) -> str:
        agent_name = context.get("agent_name", "the agent")
        return (
            dedent(
                f"""
                You are helpful AI assistant {agent_name}.
                Follow the user's instructions carefully and provide the best possible response.
                If you are unsure about any aspect of the instructions, ask for clarification.
                Always aim to be clear, concise, and informative in your responses.
                """
            ).strip()
            + "\n"
        )

    def load_scaffold(
        self, loader: Callable[[Any | None], dict[str, Any]], agent_id: int | None = None
    ) -> Any:
        return loader(agent_id)



from __future__ import annotations

import os
from dataclasses import dataclass


@dataclass(frozen=True)
class AgentHubConfig:
    base_url: str
    attp_url: str
    agt_key: str
    organization_id: int


def _get_env(name: str, default: str) -> str:
    value = os.getenv(name, default)
    return value.strip() or default


def load_agenthub_config() -> AgentHubConfig:
    base_url = _get_env("AGENTHUB__BASE_URL", "http://localhost:8000")
    attp_url = _get_env("AGENTHUB__ATTP_URL", "attp://localhost:6563")
    agt_key = _get_env("AGENTHUB__AGT_KEY", "agt_eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJvcmdhbml6YXRpb25faWQiOjEsInBlcm1pc3Npb25zIjpbIioiXSwiZXhwaXJlc19hdF90aW1lc3RhbXAiOm51bGwsInV1aWQiOiJkMjQwYzI3ZC04ODRjLTQ3Y2UtOWEyOC1jZWM3NTgzZTJmYjciLCJhZG1pbmlzdHJhdGl2ZSI6dHJ1ZX0.hRL3zgVMZqvo43IPT3ZTLixLVsiayJ618bJW4DZ3kYs")
    org_raw = _get_env("AGENTHUB__ORGANIZATION_ID", "1")
    try:
        organization_id = int(org_raw)
    except ValueError:
        organization_id = 1
    return AgentHubConfig(
        base_url=base_url,
        attp_url=attp_url,
        agt_key=agt_key,
        organization_id=organization_id,
    )



"""AgentHub copilot CLI."""

from __future__ import annotations

from pathlib import Path

from ascender.core.cli_engine import Command, GenericCLI, Handler, Parameter
from attp_client.client import ATTPClient
from attp_core.rs_api import Limits
from rich.console import Console
from rich.prompt import Confirm

from agenthub.configs.main import load_agenthub_config
from agenthub.copilot.sdk import AgentCreationCopilotSDK
from agenthub.copilot.ui import run_copilot_tui


PROJECT_ROOT = Path(__file__).resolve().parents[3]
SRC_DIR = PROJECT_ROOT / "src"


@Command(name="copilot", description="AgentHub copilot utilities")
class CopilotCLI(GenericCLI):
    def __init__(self) -> None:
        super().__init__()
        self.console = Console()
        self.scaffolds_dir = SRC_DIR / "scaffolds"

    @Handler("scaffold", description="Generate the agent-creation-copilot scaffold")
    async def generate_scaffold(
        self,
        overwrite: bool = Parameter(
            False,
            description="Overwrite existing scaffold files",
            names=["-f", "--force"],
        ),
    ) -> None:
        sdk = AgentCreationCopilotSDK()
        scaffold_path, prompt_path = sdk.scaffold_paths()

        if (scaffold_path.exists() or prompt_path.exists()) and not overwrite:
            if not Confirm.ask("Scaffold exists. Overwrite?"):
                self.console.print("[yellow]Aborted - existing files left untouched.[/yellow]")
                return

        if overwrite and (scaffold_path.exists() or prompt_path.exists()):
            if not Confirm.ask("Overwrite existing scaffold files?"):
                self.console.print("[yellow]Aborted - existing files left untouched.[/yellow]")
                return

        sdk.write_scaffold(overwrite=True)
        self.console.print(
            f"[green]Scaffold created:[/green] {scaffold_path.relative_to(PROJECT_ROOT)}"
        )
        self.console.print(
            f"[green]Prompt created:[/green] {prompt_path.relative_to(PROJECT_ROOT)}"
        )

    @Handler("run", description="Run the copilot via ATTP")
    async def run_copilot(
        self,
        subject: str | None = Parameter(
            None,
            description="Optional subject. If omitted, the most recent chat is reused.",
            names=["--subject"],
        ),
    ) -> None:
        scaffold_path = self.scaffolds_dir / "agent_creation_copilot.py"
        if not scaffold_path.exists():
            self.console.print("[red]Copilot scaffold not found.[/red]")
            self.console.print("[yellow]Generate it with 'ascender run copilot scaffold'.[/yellow]")
            return

        config = load_agenthub_config()
        client = ATTPClient(
            connection_url=config.attp_url,
            agt_token=config.agt_key,
            organization_id=config.organization_id,
            limits=Limits(max_payload_size=600_000_000),
            logger=None,
        )

        sdk = AgentCreationCopilotSDK()
        agent_name = sdk.spec.agent_name
        chat_name = f"Agent Copilot Session {subject}" if subject else None

        try:
            await client.connect()
        except Exception as exc:
            self.console.print(f"[red]Failed to connect to AgentHub:[/red] {exc}")
            return

        try:
            await run_copilot_tui(
                client=client,
                agent_name=agent_name,
                chat_name=chat_name,
            )
        finally:
            await client.close()



from typing import Any, Callable, Mapping

from pydantic import validate_call

from attp_client import AttpCatalog, FixedBaseModel, IEnvelope


class ToolKind:
    def __init__(
        self, 
        metadata: dict[str, Any],
        callback: Callable[..., Any]
    ) -> None:
        self.metadata = metadata
        self.callback = validate_call(config={"arbitrary_types_allowed": True})(callback)

    async def execute_cb(self, envelope: IEnvelope) -> Any:
        return await self.callback(data=envelope.data, metadata=envelope.metadata)
    
    async def attach(self, catalog: AttpCatalog):
        catalog.attach_tool(
            self.execute_cb,
            self.metadata["name"],
            description=self.metadata.get("description"),
            schema=self.metadata.get("schema"),
            schema_id=self.metadata.get("schema_id"),
            return_direct=self.metadata.get("return_direct", False),
            schema_ver=self.metadata.get("schema_ver", "1.0"),
            timeout_ms=self.metadata.get("timeout_ms", 20000),
            idempotent=self.metadata.get("idempotent", False)
        )
    
def tool(
    name: str, 
    description: str | None = None,
    schema: type[FixedBaseModel] | dict | None = None,
    schema_id: str | None = None,
    *,
    return_direct: bool = False,
    schema_ver: str = "2.0",
    timeout_ms: float = 20000,
    idempotent: bool = False
):
    def decorator(func):
        metadata = {
            "name": name,
            "description": description or func.__doc__ or "",
            "schema": schema.model_json_schema() if isinstance(schema, type) and issubclass(schema, FixedBaseModel) else schema,
            "schema_id": schema_id,
            "return_direct": return_direct,
            "schema_ver": schema_ver,
            "timeout_ms": timeout_ms,
            "idempotent": idempotent
        }
        return ToolKind(metadata, func)
    return decorator


"""
Auto-generated scaffold for test_common_agent.
Modify AGENT_DESCRIPTION and configuration fields as needed.
"""

from __future__ import annotations

from typing import Any

from agenthub.interfaces.scaffolds.common_agent import CommonAgentData


AGENT_NAME = 'test_common_agent'
AGENT_DESCRIPTION = 'test_common_agent agent description'

AGENT_ID = None  # Not defined if not synchronized with AgentHub.
MODULE_ID = 'common-agent'
LLM_STRING = 'openai:gpt-5'
PROMPT_FILENAME = 'test_common_agent.prompt'


def _coerce_prompt(prompt_text: str | None) -> str | None:
    if not prompt_text:
        return None

    return prompt_text.strip()


def get_agent_manifest(prompt: str | None = None) -> dict[str, Any]:
    configuration = get_agent_configuration(prompt)
    return {
        "name": AGENT_NAME,
        "avatar_url": None,
        "description": AGENT_DESCRIPTION,
        "module_id": MODULE_ID,
        "configuration": configuration,
    }


def get_agent_configuration(prompt: str | None = None) -> CommonAgentData:
    return CommonAgentData(
        llm=LLM_STRING,
        instructions=_coerce_prompt(prompt) or "'You are an AgentHub agent. Fall back to this instruction when no prompt is provided.'",
        enable_attp=True,
        enable_mcp=False,
        max_iterations=10
    )



from typing import Any
from ascender.core import Application, Provider
from fastapi_mcp import FastApiMCP


class LifeCycleHook:
    def __init__(self, application: Application, **data) -> None:
        self.instance = FastApiMCP(**data)
        self.application = application
                
    def __call__(self) -> Any:
        self.instance.mount_http()
        return self.instance


def provideMCP(
    name: str | None = None,
    description: str | None = None,
    **additional_kwargs: Any,
) -> Provider:
    return {
        "provide": FastApiMCP,
        "use_factory": lambda a: LifeCycleHook(a, name=name, description=description, **additional_kwargs),
        "deps": [Application],
    }


"""Agent scaffold management CLI."""

from __future__ import annotations

import importlib.util
import importlib.resources as resources
import inspect
import re
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, cast

from ascender.core.cli_engine import Command, GenericCLI, Handler, Parameter
from attp_client import CorrelatedRPCException, NotFoundError
from attp_client.client import ATTPClient
from attp_client.interfaces.objects.agent import IAgentDTO, IAgentResponse
from attp_core.rs_api import Limits
from rich.console import Console
from rich.prompt import Confirm, Prompt
from rich.table import Table

from agenthub.configs.main import AgentHubConfig, load_agenthub_config
from agenthub.interfaces.scaffold_module import ScaffoldModule
from agenthub.core.scaffolds import ScaffoldEngine, ScaffoldGenerationSpec


PROJECT_ROOT = Path.cwd()
SRC_DIR = PROJECT_ROOT


@dataclass(slots=True)
class AgentScaffold:
    slug: str
    name: str
    manifest_loader: Callable[[str | None], dict[str, Any]]
    path: Path
    prompt_path: Path
    module_id: str
    description: str
    agent_id: int | None

    def load_manifest(self, prompt_override: str | None) -> dict[str, Any]:
        return self.manifest_loader(prompt_override)


@dataclass(slots=True)
class ScaffoldModuleSpec:
    module_id: str
    name: str
    description: str
    instance: ScaffoldModule
    path: Path


def _slugify(value: str) -> str:
    slug = re.sub(r"[^a-z0-9]+", "_", value.lower()).strip("_")
    return slug or "agent"


def _python_repr(value: str | None) -> str:
    return "None" if value is None else repr(value)


def _load_scaffold(module_path: Path) -> AgentScaffold:
    module_name = f"scaffolds.{module_path.stem}"
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if not spec or not spec.loader:
        raise RuntimeError(f"Unable to load scaffold module from {module_path}")

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    agent_name = getattr(module, "AGENT_NAME", module_path.stem)
    agent_id = getattr(module, "AGENT_ID", None)
    manifest_loader = getattr(module, "get_agent_manifest", None)

    if not callable(manifest_loader):
        raise RuntimeError(f"Scaffold {module_path} is missing get_agent_manifest")

    manifest = cast(dict[str, Any], manifest_loader(None))
    module_id = manifest.get("module_id", "rpc-agent")
    description = manifest.get("description", "")

    return AgentScaffold(
        slug=module_path.stem,
        name=agent_name,
        manifest_loader=cast(Callable[[str | None], dict[str, Any]], manifest_loader),
        path=module_path,
        prompt_path=module_path.with_suffix(".prompt"),
        module_id=module_id,
        description=description,
        agent_id=agent_id,
    )


@Command(name="agents", description="AgentHub agent builder CLI")
class AgentBuilderCLI(GenericCLI):
    """Utility commands for working with AgentHub scaffolds."""

    def __init__(self) -> None:
        super().__init__()
        self.console = Console()
        self.project_root = PROJECT_ROOT
        self.scaffolds_dir = PROJECT_ROOT / "scaffolds"
        self.scaffold_engine = ScaffoldEngine(self.scaffolds_dir)
        self.scaffold_modules = self._discover_scaffold_modules()
        self.settings: AgentHubConfig = load_agenthub_config()

    @Handler("new", description="Create a new AgentHub agent scaffold")
    async def create_agent_scaffold(
        self,
        name: str = Parameter(description="AgentHub agent name (will be slugified for files)", names=["name"]),
        description: str | None = Parameter(
            None,
            description="Agent description",
            names=["-d", "--description"],
        ),
        module_id: str = Parameter(
            "common-agent",
            description="AgentHub module id",
            names=["-m", "--module"],
        ),
        agent_id: int | None = Parameter(
            None,
            description="Expected AgentHub ID",
            names=["-i", "--id"],
        ),
        llm_load_string: str = Parameter(
            "openai:gpt-5",
            description="LLM load string for RPCAgentData",
            names=["--llm"],
        ),
        instruction_template: str | None = Parameter(
            None,
            description="Fallback instruction template if prompt is empty",
            names=["--instruction"],
        ),
        avatar_url: str | None = Parameter(
            None,
            description="Optional avatar URL",
            names=["--avatar"],
        ),
        attp_catalog: str | None = Parameter(
            None,
            description="Optional ATTP catalog name",
            names=["--catalog"],
        ),
        enable_tables_reading: bool = Parameter(
            False,
            description="Enable tables reading mode",
            names=["--tables"],
        ),
        enable_attp: bool = Parameter(
            True,
            description="Control ATTP integration flag",
            names=["--attp"],
        ),
        overwrite: bool = Parameter(
            False,
            description="Overwrite scaffold if it already exists",
            names=["-f", "--force"],
        ),
    ) -> None:
        slug = _slugify(name)
        scaffold_path = self.scaffolds_dir / f"{slug}.py"
        prompt_path = scaffold_path.with_suffix(".prompt")

        if scaffold_path.exists() and not overwrite:
            if not Confirm.ask(f"{scaffold_path.name} exists. Overwrite?"):
                self.console.print("[yellow]Aborted - existing files left untouched.[/yellow]")
                return

        module_spec = self.scaffold_modules.get(module_id)
        if not module_spec:
            available = ", ".join(sorted(self.scaffold_modules.keys()))
            self.console.print(f"[red]Unknown module '{module_id}'.[/red]")
            if available:
                self.console.print(f"[yellow]Available modules:[/yellow] {available}")
            return

        description = description or f"{name} agent description"
        instruction_template = instruction_template or (
            "You are an AgentHub agent. Fall back to this instruction when no prompt is provided."
        )

        context = {
            "agent_name": name,
            "agent_name_repr": repr(name),
            "agent_id": agent_id,
            "description_repr": repr(description),
            "module_id": module_id,
            "module_id_repr": repr(module_id),
            "llm_load_string_repr": repr(llm_load_string),
            "prompt_filename_repr": repr(prompt_path.name),
            "instruction_template_repr": repr(instruction_template),
            "avatar_url_repr": _python_repr(avatar_url),
            "attp_catalog_repr": _python_repr(attp_catalog),
            "enable_attp": enable_attp,
            "enable_tables_reading": enable_tables_reading,
        }

        prompt_default = module_spec.instance.make_prompt_template(context)
        prompt_text = Prompt.ask("Prompt text", default=prompt_default)

        gen_spec = ScaffoldGenerationSpec(
            name=name,
            description=description,
            module_id=module_id,
            agent_id=agent_id,
            llm_load_string=llm_load_string,
            instruction_template=instruction_template,
            avatar_url=avatar_url,
            attp_catalog=attp_catalog,
            prompt_text=prompt_text,
            overwrite=overwrite,
            scaffolds_dir=self.scaffolds_dir,
        )

        self.scaffold_engine.generate_scaffold(gen_spec)

        self.console.print(
            f"[green]Scaffold created:[/green] {scaffold_path.relative_to(self.project_root)}"
        )
        self.console.print(
            f"[green]Prompt created:[/green] {prompt_path.relative_to(self.project_root)}"
        )

    @Handler("status", description="Show local scaffold state compared to AgentHub")
    async def show_status(self) -> None:
        scaffolds = self._discover_scaffolds()
        if not scaffolds:
            self.console.print("[yellow]No agent scaffolds found.[/yellow]")
            return

        table = Table(title="AgentHub Scaffold Status")
        table.add_column("Agent", style="cyan")
        table.add_column("ID", justify="right")
        table.add_column("Module")
        table.add_column("Prompt", justify="center")
        table.add_column("Remote", justify="center")
        table.add_column("Notes", style="magenta")

        client: ATTPClient | None = None
        connect_error: Exception | None = None

        try:
            client = self._build_client()
            await client.connect()
        except Exception as exc:  # pragma: no cover - connection issues are runtime dependent
            connect_error = exc

        for scaffold in scaffolds:
            dto, _ = self._build_local_agent(scaffold)
            prompt_exists = scaffold.prompt_path.exists()
            prompt_label = "yes" if prompt_exists else "no"
            remote_state = "N/A"
            remote_agent: IAgentResponse | None = None
            notes: list[str] = []

            if connect_error:
                remote_state = "[red]error[/red]"
                notes.append("AgentHub unreachable")
            elif client:
                try:
                    remote_agent, lookup_notes = await self._get_remote_agent(client, scaffold, dto)
                    notes.extend(lookup_notes)
                except Exception as exc:  # pragma: no cover
                    remote_state = "[red]error[/red]"
                    notes.append(str(exc))
                else:
                    if remote_agent is None:
                        remote_state = "[red]missing[/red]"
                        notes.append("Create with 'ascender run agents upgrade'")
                        if scaffold.agent_id is not None:
                            notes.append(f"No remote agent with id {scaffold.agent_id}")
                    else:
                        diff_fields = self._diff_agent(dto, remote_agent)
                        if diff_fields:
                            remote_state = "[yellow]update[/yellow]"
                            notes.append("Fields: " + ", ".join(diff_fields))
                        else:
                            remote_state = "[green]in sync[/green]"
                        if scaffold.agent_id and remote_agent.id != scaffold.agent_id:
                            notes.append(
                                f"ID mismatch (local {scaffold.agent_id}, remote {remote_agent.id})"
                            )
                        if scaffold.agent_id is None:
                            notes.append(f"Remote id {remote_agent.id}; consider setting AGENT_ID in scaffold")

            table.add_row(
                dto.name,
                str(remote_agent.id) if remote_agent else "-",
                dto.module_id,
                prompt_label,
                remote_state,
                "; ".join(notes) if notes else "",
            )

        if client:
            await client.close()

        if connect_error:
            self.console.print(f"[red]Failed to reach AgentHub:[/red] {connect_error}")

        self.console.print(table)

    @Handler("upgrade", description="Create missing AgentHub agents from scaffolds")
    async def upgrade_agents(
        self,
        dry_run: bool = Parameter(
            False,
            description="Only show actions without creating agents",
            names=["--dry-run"],
        ),
    ) -> None:
        scaffolds = self._discover_scaffolds()
        if not scaffolds:
            self.console.print("[yellow]No agent scaffolds found.[/yellow]")
            return

        client = self._build_client()
        try:
            await client.connect()
        except Exception as exc:
            self.console.print(f"[red]Failed to connect to AgentHub:[/red] {exc}")
            return

        created = 0
        updated = 0
        unchanged = 0
        try:
            for scaffold in scaffolds:
                dto, _ = self._build_local_agent(scaffold)
                remote_agent, lookup_notes = await self._get_remote_agent(client, scaffold, dto)

                for note in lookup_notes:
                    self.console.print(f"[yellow]{dto.name}: {note}[/yellow]")

                if remote_agent is None:
                    if dry_run:
                        self.console.print(f"[cyan]Would create agent {dto.name}[/cyan]")
                        created += 1
                        continue

                    response = await client.agents.create_agent(dto)
                    created += 1
                    self.console.print(
                        f"[green]Created agent {response.id}[/green] ({dto.name})"
                    )
                    if scaffold.agent_id and response.id != scaffold.agent_id:
                        self.console.print(
                            f"[yellow]Local AGENT_ID {scaffold.agent_id} differs from remote {response.id}; update scaffold.[/yellow]"
                        )
                    if scaffold.agent_id is None:
                        self.console.print(
                            f"[yellow]Set AGENT_ID={response.id} in scaffold for future sync.[/yellow]"
                        )
                    continue

                diff_fields = self._diff_agent(dto, remote_agent)
                if not diff_fields:
                    unchanged += 1
                    if dry_run:
                        self.console.print(f"[blue]{dto.name} is up to date.[/blue]")
                    continue

                fields_label = ", ".join(diff_fields)
                if dry_run:
                    self.console.print(
                        f"[cyan]Would update agent {remote_agent.id} ({dto.name}) fields: {fields_label}"
                    )
                    updated += 1
                    continue

                await client.agents.update_agent(remote_agent.id, dto)
                updated += 1
                self.console.print(
                    f"[green]Updated agent {remote_agent.id}[/green] ({dto.name}) fields: {fields_label}"
                )
                if scaffold.agent_id and remote_agent.id != scaffold.agent_id:
                    self.console.print(
                        f"[yellow]Local AGENT_ID {scaffold.agent_id} differs from remote {remote_agent.id}; update scaffold.[/yellow]"
                    )
                if scaffold.agent_id is None:
                    self.console.print(
                        f"[yellow]Remote id {remote_agent.id}; consider setting AGENT_ID in scaffold.[/yellow]"
                    )
        finally:
            await client.close()

        if dry_run:
            self.console.print(
                f"[blue]Dry run complete: create {created}, update {updated}, unchanged {unchanged}.[/blue]"
            )
        else:
            self.console.print(
                f"[green]Done. Created {created}, updated {updated}, unchanged {unchanged}.[/green]"
            )

    def _build_local_agent(self, scaffold: AgentScaffold) -> tuple[IAgentDTO, dict[str, Any]]:
        prompt_text = (
            scaffold.prompt_path.read_text(encoding="utf-8")
            if scaffold.prompt_path.exists()
            else None
        )
        manifest = scaffold.load_manifest(prompt_text)

        configuration = manifest.get("configuration")
        if hasattr(configuration, "model_dump"):
            configuration = configuration.model_dump(mode="json")  # type: ignore

        dto = IAgentDTO(
            name=manifest.get("name", scaffold.name),
            description=manifest.get("description", scaffold.description),
            module_id=manifest.get("module_id", scaffold.module_id),
            avatar_url=manifest.get("avatar_url"),
            configurations=configuration or {},
        )

        return dto, manifest

    async def _get_remote_agent(
        self,
        client: ATTPClient,
        scaffold: AgentScaffold,
        dto: IAgentDTO,
    ) -> tuple[IAgentResponse | None, list[str]]:
        try:
            remote = await client.agents.get_agent(agent_name=dto.name)
            return remote, []
        except (NotFoundError, CorrelatedRPCException):
            remote = None

        if scaffold.agent_id is not None:
            try:
                remote = await client.agents.get_agent(agent_id=scaffold.agent_id)
            except NotFoundError:
                return None, []

            notes: list[str] = []
            if remote.name != dto.name:
                notes.append(f"Will rename from '{remote.name}' to '{dto.name}'")
            return remote, notes

        return None, []

    @staticmethod
    def _diff_agent(dto: IAgentDTO, remote: IAgentResponse) -> list[str]:
        local = dto.model_dump(mode="json")
        remote_dump = remote.model_dump(mode="json")
        diffs: list[str] = []

        for key, value in local.items():
            remote_value = remote_dump.get(key)
            if key == "configurations":
                remote_value = remote_value or {}
                value = value or {}
            if remote_value != value:
                diffs.append(key)

        return diffs

    def _discover_scaffolds(self) -> list[AgentScaffold]:
        scaffolds: list[AgentScaffold] = []
        for path in sorted(self.scaffolds_dir.glob("*.py")):
            try:
                scaffolds.append(_load_scaffold(path))
            except Exception as exc:
                self.console.print(f"[red]Failed to load {path.name}:[/red] {exc}")
        return scaffolds

    def _discover_scaffold_modules(self) -> dict[str, ScaffoldModuleSpec]:
        specs: dict[str, ScaffoldModuleSpec] = {}
        for module in self.scaffold_engine.discover_scaffold_modules().values():
            specs[module.module_id] = ScaffoldModuleSpec(
                module_id=module.module_id,
                name=module.name,
                description=module.description,
                instance=module,
                path=Path("<package>"),
            )
        return specs

    def _build_client(self) -> ATTPClient:
        return ATTPClient(
            connection_url=self.settings.attp_url,
            agt_token=self.settings.agt_key,
            organization_id=self.settings.organization_id,
            limits=Limits(max_payload_size=600_000_000),
            logger=None,
        )



import inspect
import traceback
from typing import Annotated, Any, get_args, get_origin

from ascender.common import BaseDTO, BaseResponse
from ascender.core import Application, ControllerDecoratorHook, inject
from attp_client import ATTPClient, AttpException, FixedBaseModel
from attp_client.interfaces.catalogs.tools.envelope import IEnvelope
from pydantic import BaseModel

from .context import CtxMarker
from .pipe import Pipe
from ..interfaces.types.tool_metadata import ToolMetadata


class AttpTool(ControllerDecoratorHook):
    """
    A controller decorator hook for ATTP tool integration.
    
    It registers ATTP Tool in scope of selected category provided in the decorator parameters.
    It executes the callback function as soon as AgentHub Agent calls the tool correlated with this decorator.
    """
    attp_client: ATTPClient = inject(ATTPClient)
    application: Application = inject(Application)
    
    def __init__(
        self, 
        name: str,
        description: str | None = None,
        catalog: str = "default"
    ) -> None:
        super().__init__()
        
        self.name = name
        self.description = description
        self.catalog = catalog

    def __resolve_fixed_model(self, annotation: Any):
        """Return FixedBaseModel subclass if present in annotation."""
        if inspect.isclass(annotation) and issubclass(annotation, (FixedBaseModel, BaseModel, BaseDTO, BaseResponse)):
            return annotation

        origin = get_origin(annotation)
        if inspect.isclass(origin) and issubclass(origin, (FixedBaseModel, BaseModel, BaseDTO, BaseResponse)):
            return origin

        return None

    def __is_context_param(self, param: inspect.Parameter) -> bool:
        if isinstance(param.default, CtxMarker):
            return True

        annotation = param.annotation
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is Annotated and any(isinstance(arg, CtxMarker) for arg in args[1:]):
            return True

        return False

    def __is_pipe_param(self, param: inspect.Parameter) -> bool:
        annotation = param.annotation
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is Annotated and any(isinstance(arg, Pipe) for arg in args[1:]):
            return True

        return False
    
    def get_pipe_param(self, param: inspect.Parameter) -> Pipe | None:
        annotation = param.annotation
        origin = get_origin(annotation)
        args = get_args(annotation)
        if origin is Annotated:
            for arg in args[1:]:
                if isinstance(arg, Pipe):
                    return arg
        return None

    def __extract_schema(self, callback: Any):
        """Extract schema from the callback's annotations if it uses FixedBaseModel."""
        for param in inspect.signature(callback).parameters.values():
            if self.__is_context_param(param):
                continue
            
            if self.__is_pipe_param(param):
                continue

            annotation = param.annotation
            model = self.__resolve_fixed_model(annotation)
            if model:
                return model.model_json_schema()

        return self.__build_schema_from_signature(callback)

    def __map_type_to_schema(self, annotation: Any) -> dict[str, Any]:
        origin = get_origin(annotation) or annotation
        args = get_args(annotation)

        if origin in (str, bytes):
            return {"type": "string"}
        if origin is bool:
            return {"type": "boolean"}
        if origin is int:
            return {"type": "integer"}
        if origin is float:
            return {"type": "number"}
        if origin is dict:
            return {"type": "object"}
        if origin in (list, tuple, set):
            items_schema = self.__map_type_to_schema(args[0]) if args else {}
            return {"type": "array", "items": items_schema}
        # Fallback to permissive schema for unknown types
        return {}

    def __build_schema_from_signature(self, callback: Any) -> dict[str, Any] | None:
        sig = inspect.signature(callback)
        properties: dict[str, Any] = {}
        required: list[str] = []

        for name, param in sig.parameters.items():
            if name in ("self", "cls"):
                continue

            if self.__is_context_param(param):
                continue

            annotation = param.annotation
            args = get_args(annotation)
            base_annotation = next((arg for arg in args if arg is not type(None)), annotation)

            schema = self.__map_type_to_schema(base_annotation)
            properties[name] = schema

            default = param.default
            allows_none = type(None) in args
            if default is inspect._empty and not allows_none:
                required.append(name)

        if not properties:
            return None

        return {
            "type": "object",
            "properties": properties,
            "required": required
        }

    def __unwrap_annotation(self, annotation: Any) -> Any:
        origin = get_origin(annotation)
        if origin is Annotated:
            return get_args(annotation)[0]
        if args := get_args(annotation):
            return next((arg for arg in args if arg is not type(None)), annotation)
        return annotation

    async def __build_kwargs_from_envelope(self, envelope: IEnvelope, callback: Any) -> dict[str, Any]:
        sig = inspect.signature(callback)
        data = envelope.data or {}
        kwargs: dict[str, Any] = {}
        model_params: dict[str, Any] = {}

        for name, param in sig.parameters.items():
            if name in ("self", "cls"):
                continue

            if self.__is_context_param(param) or self.__is_pipe_param(param):
                continue

            annotation = self.__unwrap_annotation(param.annotation)
            model = self.__resolve_fixed_model(annotation)
            if model:
                model_params[name] = model

        for name, param in sig.parameters.items():
            if name in ("self", "cls"):
                continue

            if self.__is_context_param(param):
                metadata: ToolMetadata = envelope.metadata or {} # type: ignore
                kwargs[name] = metadata
                continue
            
            if self.__is_pipe_param(param):
                pipe = self.get_pipe_param(param)
                if not pipe:
                    continue
                supply = pipe.supply(envelope)
                
                if inspect.iscoroutine(supply):
                    kwargs[name] = await supply
                    continue
                
                kwargs[name] = supply
                continue

            if name in model_params:
                kwargs[name] = model_params[name].model_validate(data)
                continue

            if name not in data and param.default is inspect._empty:
                raise ValueError(f"Missing required tool argument '{name}'")

            if name not in data:
                continue  # default will be used by Python

            annotation = self.__unwrap_annotation(param.annotation)
            model = self.__resolve_fixed_model(annotation)
            value = data[name]
                
            kwargs[name] = model.model_validate(value) if model else value

        return kwargs

    def on_load(self, callback):
        schema = self.__extract_schema(callback)

        async def _invoke(envelope: IEnvelope):
            try:
                kwargs = await self.__build_kwargs_from_envelope(envelope, callback)
                return await callback(**kwargs)
            except Exception as e:
                traceback.print_exc()
                raise AttpException(detail={"message": f"Error invoking ATTP tool '{self.name}': {e}"}) from e

        async def _attach():
            catalog = await self.attp_client.catalog(self.catalog)
            catalog.attach_tool(
                _invoke,
                self.name,
                description=self.description or callback.__doc__,
                schema=schema,
                schema_id=self.name + "_schema",
                schema_ver="2.0"
            )
        
        async def _detach():
            print(f"Detatching tool from ATTP... {self.name}")
            catalog = await self.attp_client.catalog(self.catalog)
            catalog.detach_tool(self.name)

        self.application.app.add_event_handler("startup", _attach)
        self.application.app.add_event_handler("shutdown", _detach)



from __future__ import annotations

import re
from dataclasses import dataclass, field
from uuid import UUID
from typing import cast

from attp_client.client import ATTPClient
from attp_client.interfaces.inference.enums.message_type import MessageTypeEnum
from attp_client.interfaces.inference.message import IMessageDTOV2
from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout import HSplit, Layout, VSplit
from prompt_toolkit.layout.containers import Window
from prompt_toolkit.layout.controls import FormattedTextControl
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.widgets import Frame, TextArea

from agenthub.copilot.tools import CopilotTools


@dataclass(frozen=True)
class _ChatEntry:
    id: UUID
    name: str


@dataclass
class _CopilotSession:
    name: str
    chat_id: UUID


@dataclass
class _CopilotState:
    mode: str = "roam"
    chats: list[_ChatEntry] = field(default_factory=list)
    selected_index: int = 0
    session: _CopilotSession | None = None
    messages: dict[UUID, list[tuple[str, str]]] = field(default_factory=dict)

    def current_chat_id(self) -> UUID | None:
        return self.session.chat_id if self.session else None

    def current_messages(self) -> list[tuple[str, str]]:
        chat_id = self.current_chat_id()
        if not chat_id:
            return []
        return self.messages.setdefault(chat_id, [])


class _CopilotTUI:
    def __init__(self, state: _CopilotState) -> None:
        self.state = state
        self.sidebar = TextArea(read_only=True, scrollbar=True)
        self.chat_area = TextArea(read_only=True, scrollbar=True)
        self.input_area = TextArea(height=5, multiline=True)
        self.status_control = FormattedTextControl(text="")

        self.sidebar_frame = Frame(self.sidebar, title="Copilot Sessions")
        self.chat_frame = Frame(self.chat_area, title="Agent Copilot")
        self.input_frame = Frame(self.input_area, title="Input")

        self.root = VSplit(
            [
                self.sidebar_frame,
                HSplit(
                    [
                        self.chat_frame,
                        self.input_frame,
                        Window(height=1, content=self.status_control),
                    ],
                    padding=0,
                ),
            ],
            padding=1,
            width=None,
        )

    def build_layout(self) -> Layout:
        return Layout(self.root, focused_element=self.sidebar)

    def set_mode(self, mode: str) -> None:
        self.state.mode = mode
        if mode == "insert":
            self.input_area.read_only = False
        else:
            self.input_area.read_only = True

    def update_sidebar(self) -> None:
        lines: list[str] = []
        if not self.state.chats:
            lines.append("No chats found.")
        else:
            for idx, chat in enumerate(self.state.chats):
                prefix = ">" if idx == self.state.selected_index else " "
                lines.append(f"{prefix} {idx + 1:>2}. {chat.name}")
        lines.append("")
        lines.append("Esc: roam | Enter: insert")
        lines.append("Tab: next pane | /exit to quit")
        lines.append("/refresh, /select <n>, /new <name>")
        self.sidebar.text = "\n".join(lines)

    def update_chat(self) -> None:
        lines: list[str] = []
        for role, message in self.state.current_messages()[-300:]:
            lines.append(f"{role}: {message}")
        if not lines:
            lines.append("Start the conversation below.")
        self.chat_area.text = "\n".join(lines)
        # Auto-scroll to the bottom for the latest messages.
        self.chat_area.buffer.cursor_position = len(self.chat_area.text)

    def update_status(self) -> None:
        active = self.state.session.name if self.state.session else "none"
        self.status_control.text = f"mode={self.state.mode} | chat={active}"

    def refresh(self) -> None:
        self.update_sidebar()
        self.update_chat()
        self.update_status()


def _append_tool_messages(
    messages: list[tuple[str, str]], response: IMessageDTOV2
) -> None:
    if response.tool_called:
        messages.append(("system", f"[tool] {response.tool_called.name}"))
    if response.tool_status:
        messages.append(("system", f"[tool-status] {response.tool_status}"))
    if response.tool_started_input:
        messages.append(("system", f"[tool-input] {response.tool_started_input}"))
    if response.tool_finished_output:
        messages.append(("system", f"[tool-output] {response.tool_finished_output}"))
    if response.tool_error_detail:
        messages.append(("system", f"[tool-error] {response.tool_error_detail}"))


async def run_copilot_tui(
    client: ATTPClient,
    agent_name: str,
    chat_name: str | None = None,
) -> None:
    tools = CopilotTools(client)
    await tools.attach_all()

    async def _fetch_chats() -> list[_ChatEntry]:
        try:
            response = await client.chats.get_chats()
            return [
                _ChatEntry(id=getattr(chat, "id"), name=getattr(chat, "name", ""))
                for chat in response.items
            ]  # type: ignore[arg-type]
        except Exception:
            return []

    async def _create_chat(name: str) -> _ChatEntry:
        chat_response = await client.inference.create_chat(
            name=name,
            agent_name=agent_name,
        )
        return _ChatEntry(id=UUID(str(chat_response["id"])), name=name)

    chats = await _fetch_chats()

    selected_chat: _ChatEntry | None = None
    if chat_name:
        selected_chat = await _create_chat(chat_name)
        chats = [c for c in chats if c.id != selected_chat.id]
        chats.insert(0, selected_chat)
    else:
        if chats:
            selected_chat = chats[0]
        else:
            selected_chat = await _create_chat("Agent Copilot Session")
            chats.append(selected_chat)

    chat_id = selected_chat.id
    selected_index = next((idx for idx, c in enumerate(chats) if c.id == chat_id), 0)

    state = _CopilotState(
        chats=chats,
        selected_index=selected_index,
        session=_CopilotSession(name=selected_chat.name, chat_id=chat_id),
    )
    ui = _CopilotTUI(state)
    ui.set_mode("roam")
    ui.refresh()

    async def _switch_session(chat: _ChatEntry) -> None:
        state.session = _CopilotSession(name=chat.name, chat_id=chat.id)
        state.selected_index = next((idx for idx, c in enumerate(state.chats) if c.id == chat.id), 0)
        state.messages.setdefault(chat.id, [])
        ui.refresh()

    async def _refresh_chats(preserve_selection: bool = True) -> None:
        fetched = await _fetch_chats()
        if fetched:
            state.chats = fetched
        if preserve_selection and state.session and not any(c.id == state.session.chat_id for c in state.chats):
            state.chats.insert(0, _ChatEntry(id=state.session.chat_id, name=state.session.name))
        if preserve_selection and state.session:
            for idx, chat in enumerate(state.chats):
                if chat.id == state.session.chat_id:
                    state.selected_index = idx
                    break
        ui.refresh()

    async def _create_and_switch(name: str) -> None:
        new_chat = await _create_chat(name)
        state.chats = [c for c in state.chats if c.id != new_chat.id]
        state.chats.insert(0, new_chat)
        await _switch_session(new_chat)

    async def _handle_command(text: str) -> bool:
        lower = text.lower()
        if lower == "/refresh":
            await _refresh_chats()
            return True
        if lower.startswith("/select"):
            match = re.match(r"/select\s+(\d+)", text)
            if match:
                idx = int(match.group(1)) - 1
                if 0 <= idx < len(state.chats):
                    await _switch_session(state.chats[idx])
            return True
        if lower.startswith("/new"):
            name = text.split(" ", 1)[1].strip() if " " in text else "Agent Copilot Session"
            await _create_and_switch(name or "Agent Copilot Session")
            return True
        return False

    async def _stream_response(app, user_text: str) -> None:
        if not state.session:
            return

        message = IMessageDTOV2(
            content=user_text,
            message_type=MessageTypeEnum.USER_MESSAGE,
            chat_id=state.session.chat_id,
        )

        try:
            response = await client.inference.invoke_chat_inference(
                messages=[message],
                chat_id=state.session.chat_id,
                timeout=20000,
                stream=True,
            )

            if hasattr(response, "__aiter__"):
                state.current_messages().append(("copilot", ""))
                reply_index = len(state.current_messages()) - 1

                async for chunk in response:  # type: ignore[operator]
                    _append_tool_messages(state.current_messages(), chunk)
                    content = chunk.content if isinstance(chunk.content, str) else str(chunk.content or "")
                    if content:
                        role, existing = state.current_messages()[reply_index]
                        state.current_messages()[reply_index] = (role, existing + content)
                    ui.refresh()
                    app.invalidate()
                return

            resp_msg = cast(IMessageDTOV2, response)
            _append_tool_messages(state.current_messages(), resp_msg)
            reply = resp_msg.content if isinstance(resp_msg.content, str) else str(resp_msg.content)
            state.current_messages().append(("copilot", reply))
        except Exception as exc:
            state.current_messages().append(("copilot", f"Error: {exc}"))
        finally:
            ui.refresh()
            app.invalidate()

    kb = KeyBindings()

    @kb.add("escape")
    def _to_roam(event) -> None:
        ui.set_mode("roam")
        event.app.layout.focus(ui.sidebar)
        ui.refresh()
        event.app.invalidate()

    @kb.add("c-c")
    @kb.add("c-q")
    def _quit(event) -> None:
        event.app.exit()

    @kb.add("tab")
    def _focus_next(event) -> None:
        event.app.layout.focus_next()

    @kb.add("s-tab")
    def _focus_prev(event) -> None:
        event.app.layout.focus_previous()

    @kb.add("up")
    def _move_up(event) -> None:
        if event.app.layout.current_control == ui.sidebar.control:
            state.selected_index = max(0, state.selected_index - 1)
            ui.refresh()
            event.app.invalidate()

    @kb.add("down")
    def _move_down(event) -> None:
        if event.app.layout.current_control == ui.sidebar.control:
            state.selected_index = min(len(state.chats) - 1, state.selected_index + 1)
            ui.refresh()
            event.app.invalidate()

    @kb.add("enter")
    async def _enter(event) -> None:
        if state.mode == "insert":
            text = ui.input_area.text.strip()
            if not text:
                return
            if text.lower() in {"/exit", "/quit"}:
                event.app.exit()
                return
            if text.startswith("/"):
                if await _handle_command(text):
                    ui.input_area.text = ""
                    ui.refresh()
                    event.app.invalidate()
                return
            if not state.session:
                state.current_messages().append(("system", "No chat selected."))
                ui.refresh()
                event.app.invalidate()
                return
            ui.input_area.text = ""
            state.current_messages().append(("you", text))
            ui.refresh()
            event.app.invalidate()
            await _stream_response(event.app, text)
            return

        if event.app.layout.current_control == ui.sidebar.control:
            if 0 <= state.selected_index < len(state.chats):
                selected = state.chats[state.selected_index]
                await _switch_session(selected)
                event.app.invalidate()
                return

        ui.set_mode("insert")
        event.app.layout.focus(ui.input_area)
        ui.refresh()
        event.app.invalidate()

    @kb.add("c-j")
    def _insert_newline(event) -> None:
        if state.mode == "insert":
            ui.input_area.buffer.insert_text("\n")

    app = Application(
        layout=ui.build_layout(),
        key_bindings=kb,
        full_screen=True,
        mouse_support=True,
    )

    ui.refresh()
    with patch_stdout(raw=True):
        await app.run_async()



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent
from importlib import resources


# When used as a library, write scaffolds into the caller's working directory.
SCAFFOLDS_DIR = Path.cwd() / "scaffolds"
TEMPLATES_PACKAGE = "agenthub.copilot.templates"


@dataclass(frozen=True)
class CopilotScaffoldSpec:
    agent_name: str = "agent-creation-copilot"
    description: str = "agent-creation-copilot agent description"
    agent_id: int | None = 4
    module_id: str = "autonomous-agent"
    llm_string: str = "openai:gpt-5.1"
    prompt_filename: str = "agent_creation_copilot.prompt"


class AgentCreationCopilotSDK:
    def __init__(self, spec: CopilotScaffoldSpec | None = None) -> None:
        self.spec = spec or CopilotScaffoldSpec()

    def render_scaffold(self) -> str:
        agent_id_repr = "None" if self.spec.agent_id is None else str(self.spec.agent_id)
        return (
            dedent(
                f"""
                \"\"\"
                Auto-generated scaffold for {self.spec.agent_name}.
                Modify AGENT_DESCRIPTION and configuration fields as needed.
                \"\"\"

                from __future__ import annotations

                from typing import Any

                from agenthub.interfaces.scaffolds.autonomous_agent import (
                    AutonomousAgentData,
                    IAgentInstructionDTO,
                )


                AGENT_NAME = {self.spec.agent_name!r}
                AGENT_DESCRIPTION = {self.spec.description!r}

                AGENT_ID = {agent_id_repr}  # Not defined if not synchronized with AgentHub.
                MODULE_ID = {self.spec.module_id!r}
                LLM_STRING = {self.spec.llm_string!r}
                PROMPT_FILENAME = {self.spec.prompt_filename!r}


                def _coerce_prompt(prompt_text: str | None) -> IAgentInstructionDTO | None:
                    if not prompt_text:
                        return None

                    cleaned = "\\n".join(
                        line for line in prompt_text.splitlines()
                        if not line.lstrip().startswith("//")
                    )
                    return IAgentInstructionDTO.from_yaml(cleaned)


                def get_agent_manifest(prompt: str | None = None) -> dict[str, Any]:
                    configuration = get_agent_configuration(prompt)
                    return {{
                        "name": AGENT_NAME,
                        "avatar_url": None,
                        "description": AGENT_DESCRIPTION,
                        "module_id": MODULE_ID,
                        "configuration": configuration,
                    }}


                def get_agent_configuration(prompt: str | None = None) -> AutonomousAgentData:
                    instructions = _coerce_prompt(prompt) or IAgentInstructionDTO(
                        objective="You are an AgentHub agent. Fall back to this instruction when no prompt is provided.",
                        role=None,
                        backstory=None,
                        constraints=[],
                        preferences=[],
                        output_format="json",
                    )
                    return AutonomousAgentData(
                        llm=LLM_STRING,
                        instructions=instructions,
                        policies=["think_before_acting", "see_own_tokens"],
                    )
                """
            ).strip()
            + "\n"
        )

    def render_prompt(self) -> str:
        prompt_path = resources.files(TEMPLATES_PACKAGE) / self.spec.prompt_filename
        return prompt_path.read_text(encoding="utf-8")

    def scaffold_paths(self) -> tuple[Path, Path]:
        scaffold_path = SCAFFOLDS_DIR / "agent_creation_copilot.py"
        prompt_path = scaffold_path.with_suffix(".prompt")
        return scaffold_path, prompt_path

    def write_scaffold(self, overwrite: bool = False) -> tuple[Path, Path]:
        scaffold_path, prompt_path = self.scaffold_paths()
        if (scaffold_path.exists() or prompt_path.exists()) and not overwrite:
            raise FileExistsError("Scaffold already exists. Use overwrite=True to replace it.")

        SCAFFOLDS_DIR.mkdir(parents=True, exist_ok=True)
        scaffold_path.write_text(self.render_scaffold().rstrip() + "\n", encoding="utf-8")
        prompt_path.write_text(self.render_prompt().rstrip() + "\n", encoding="utf-8")
        return scaffold_path, prompt_path



from typing import Any
from uuid import UUID
from ascender.common import BaseDTO
from pydantic import Field


class CommonAgentData(BaseDTO):
    llm: str = Field("openai:gpt-5.1") # Design: 'provider:model' or 'openrouter:auto' for automatic selection
    invoke_configs: dict[str, Any] = Field(default_factory=dict) # Additional configurations for LLM invocation
    instructions: str = Field(..., description="High-level instructions guiding the agent's behavior (basically, its system prompt.)")
    max_iterations: int = Field(10, description="Maximum number of iterations the agent can perform")
    
    enable_attp: bool = Field(True, description="Flag to enable or disable ATTP tool usage")
    enable_mcp: bool = Field(False, description="Flag to enable or disable MCP tool usage")
    
    attp_catalogs: list[str] = Field(default_factory=list, description="List of ATTP tool catalogs the agent can use")
    mcp_catalogs: list[str] = Field(default_factory=list, description="List of MCP tool catalogs the agent can use")


from ascender.common import BaseDTO


from typing import Any
from uuid import UUID
from ascender.common import BaseDTO
from pydantic import Field


class IAgentInstructionDTO(BaseDTO):
    objective: str = Field(..., description="Primary objective or goal for the autonomous agent to achieve.")
    role: str | None = Field(None, description="Role or persona the agent should adopt while interacting.")
    backstory: str | None = Field(None, description="Background story or context for the agent's behavior.")
    constraints: list[str] = Field(default_factory=list, description="List of constraints or rules the agent must adhere to. Hard constraints that must not be violated.")
    preferences: list[str] = Field(default_factory=list, description="List of preferences or guidelines for the agent's actions. Soft preferences that can be considered but may be overridden if necessary.")
    output_format: str | None = Field(None, description="Expected output format (json, markdown, bullets, etc.) for the agent's responses.")
    
    @staticmethod
    def from_yaml(yaml_str: str) -> 'IAgentInstructionDTO':
        import yaml
        data = yaml.safe_load(yaml_str)
        return IAgentInstructionDTO(**data)


class AutonomousAgentData(BaseDTO):
    llm: str = Field("openai:gpt-5.1") # Design: 'provider:model' or 'openrouter:auto' for automatic selection
    invoke_configs: dict[str, Any] = Field(default_factory=dict) # Additional configurations for LLM invocation
    allowed_states: list[str] = Field(default_factory=lambda: ["thinking", "acting"]) # e.g. ['thinking', 'acting']
    policies: list[str] = Field(default_factory=lambda: ["think_before_acting", "minimize_tool_calls", "see_own_tokens", "prefer_fast_llm", "avoid_expensive_llm"]) # e.g. ['think_before_acting', 'minimize_tool_calls', 'see_own_tokens', 'prefer_fast_llm', 'avoid_expensive_llm']
    enabled_filters: list[str] = Field(default_factory=lambda: ["empty_response_regenerate"]) # e.g. ['profanity_filter', 'safety_filter', 'empty_response_regenerate]
    
    instructions: IAgentInstructionDTO = Field(...) # Instructions for the autonomous agent
    
    enable_attp: bool = True # Enable ATTP protocol for tool calls
    enable_mcp: bool = True  # Enable MCP protocol for tool calls
    enable_http: bool = False # Enable HTTP protocol for tool calls
    
    attp_catalogs: list[str] = Field(default_factory=list) # List of ATTP catalogs available to the agent
    mcp_catalogs: list[str] = Field(default_factory=list)  # List of MCP catalogs available to the agent
    
    tool_execution_error: str = "An error {error} occurred while trying to use the tool. Please try again later." # Error message template for tool execution errors
    
    sign_metadata: bool = True # Sign LLM response message with metadata (containing state and model that was used)


from __future__ import annotations

from textwrap import dedent
from typing import Any, Callable

from agenthub.interfaces.scaffold_module import ScaffoldModule


class AutonomousAgentScaffold(ScaffoldModule):
    module_id = "autonomous-agent"
    name = "Autonomous Agent"
    description = "Structured YAML prompt for autonomous agent instructions."

    def generate_manifest(self, context: dict[str, Any]) -> str:
        agent_id = context.get("agent_id")
        agent_id_repr = "None" if agent_id is None else str(agent_id)
        agent_name_repr = context["agent_name_repr"]
        description_repr = context["description_repr"]
        module_id_repr = context["module_id_repr"]
        llm_load_string_repr = context["llm_load_string_repr"]
        prompt_filename_repr = context["prompt_filename_repr"]
        instruction_template_repr = context["instruction_template_repr"]
        avatar_url_repr = context["avatar_url_repr"]

        return (
            dedent(
                f"""
                \"\"\"
                Auto-generated scaffold for {context['agent_name']}.
                Modify AGENT_DESCRIPTION and configuration fields as needed.
                \"\"\"

                from __future__ import annotations

                from typing import Any

                from agenthub.interfaces.scaffolds.autonomous_agent import (
                    AutonomousAgentData,
                    IAgentInstructionDTO,
                )


                AGENT_NAME = {agent_name_repr}
                AGENT_DESCRIPTION = {description_repr}

                AGENT_ID = {agent_id_repr}  # Not defined if not synchronized with AgentHub.
                MODULE_ID = {module_id_repr}
                LLM_STRING = {llm_load_string_repr}
                PROMPT_FILENAME = {prompt_filename_repr}


                def _coerce_prompt(prompt_text: str | None) -> IAgentInstructionDTO | None:
                    if not prompt_text:
                        return None

                    cleaned = \"\\n\".join(
                        line for line in prompt_text.splitlines()
                        if not line.lstrip().startswith(\"//\")
                    )
                    return IAgentInstructionDTO.from_yaml(cleaned)


                def get_agent_manifest(prompt: str | None = None) -> dict[str, Any]:
                    configuration = get_agent_configuration(prompt)
                    return {{
                        \"name\": AGENT_NAME,
                        \"avatar_url\": {avatar_url_repr},
                        \"description\": AGENT_DESCRIPTION,
                        \"module_id\": MODULE_ID,
                        \"configuration\": configuration,
                    }}


                def get_agent_configuration(prompt: str | None = None) -> AutonomousAgentData:
                    instructions = _coerce_prompt(prompt) or IAgentInstructionDTO(
                        objective={instruction_template_repr},
                        role=None,
                        backstory=None,
                        constraints=[],
                        preferences=[]
                    )
                    return AutonomousAgentData(
                        llm=LLM_STRING,
                        instructions=instructions,
                    )
                """
            ).strip()
            + "\n"
        )

    def make_prompt_template(self, context: dict[str, Any]) -> str:
        agent_name = context.get("agent_name", "the agent")
        return (
            dedent(
                f"""
                # YAML
                objective: \"Describe the main objective for {agent_name}.\"
                role: \"\"
                backstory: \"\"
                constraints:
                  - \"\"
                preferences:
                  - \"\"
                """
            ).strip()
            + "\n"
        )

    def load_scaffold(
        self, loader: Callable[[Any | None], dict[str, Any]], agent_id: int | None = None
    ) -> Any:
        return loader(agent_id)



from abc import ABC, abstractmethod
from typing import Any

from attp_client import IEnvelope


class Pipe(ABC):
    """
    > One wise man said: No message walks alone; even the most anonymous words drag their shadows of paper, language, and form are tiny traces that reveal how they came to be.
    
    So far pipes are designed to avoid repitative code by supplying commonly used context data from the envelope metadata.
    Design pipe once - use everywhere.
    
    Envelope contains crucial context information of:
    - Where tool is being called (which chat)
    - What agent is calling the tool (ID of the agent in agenthub)
    - In what catalog the tool is registered (catalog name)
    
    The most common use is `chat_id` from metadata to fetch user and chat information from the database.
    Every backend has its unique design even if they copy. But when they use AgentHub, they share the one common thing - chat_id.
    So pipes help people to design and automate the extraction of such common context data and conversion to whatever they need for their system.
    """
    ...
    
    @abstractmethod
    def supply(self, envelope: IEnvelope) -> Any:
        ...
    
    async def asupply(self, envelope: IEnvelope) -> Any:
        raise NotImplementedError("Asynchronous supply is not implemented for this pipe.")